﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Xml.Serialization;
using Assets;
using UnityEditor;
using UnityEngine;

/// <summary>
/// This script exists in ~\Editor\ and so will be run on 
/// asset import (or re-import). This will hook up the mesh 
/// to a ragdoll using the info file in the same directory
/// </summary>
public class ImportWaveformFBX : AssetPostprocessor
{
    public void OnPostprocessModel(GameObject gameObject)
    {
        // Only operate on FBX files
        if (!assetPath.EndsWith(".fbx"))
            return;

        var debugLogHistory = new HashSet<string>();

        ImportRagdoll(gameObject, assetPath, debugLogHistory);
        SetupMaterials(gameObject, assetPath, assetImporter as ModelImporter);
    }

    private static void ImportRagdoll(GameObject gameObject, string assetPath, HashSet<string> debugLogHistory)
    {
        // check we have a .ragdoll file

        string dir = Path.Combine(Path.GetDirectoryName(assetPath), "Ragdoll");
        string name = Path.GetFileNameWithoutExtension(assetPath).ToLower();
        string ragdollFilepath = Path.Combine(dir, name + ".ragdoll");

        if (!File.Exists(ragdollFilepath))
            return;

        // try to load the file
        var ragdollFile = Utils.LoadXml<RagdollFile>(ragdollFilepath);
        if (ragdollFile == null)
            return;

        var ragdollBones = ragdollFile.GetLookup();
        var ragdollBodies = new Dictionary<RagdollBone, Rigidbody>();

        bool isKinematic = true;

        foreach (Transform bone in Utils.GetBones(gameObject))
        {
            // try get ragdoll info for bone
            RagdollBone ragdollBone;
            if (!ragdollBones.TryGetValue(bone.name, out ragdollBone))
                continue;

            // add rigidbody component if object is animated
            var rigidBody = Specific_IsStatic(name, bone.name, debugLogHistory) ? null : bone.gameObject.AddComponent<Rigidbody>();

            // we add an empty, so can rotate collider to the correct orientation
            var colliderContainer = new GameObject("Collider Container");
            colliderContainer.transform.parent = bone;
            ragdollBone.AddCollider(colliderContainer, rigidBody, dir);

            if (rigidBody != null)
            {
                if (Specific_BecomeRagdollOnCollision(name, debugLogHistory))
                    bone.gameObject.AddComponent<AnimatedModelToRagdoll>();

                if (!Specific_SetRigidBodyProperties(rigidBody, name, bone.name))
                    isKinematic = false;

                // cache
                ragdollBodies.Add(ragdollBone, rigidBody);
            }
        }

        // add joints
        foreach (var bone in ragdollFile.RagdollBones)
        {
            // get parent bone

            if (string.IsNullOrEmpty(bone.ParentBoneName))
                continue;

            RagdollBone parentBone;
            if (!ragdollBones.TryGetValue(bone.ParentBoneName, out parentBone))
                continue;

            // get rigid bodies

            Rigidbody child;
            if (!ragdollBodies.TryGetValue(bone, out child))
                continue;

            Rigidbody parent;
            if (!ragdollBodies.TryGetValue(parentBone, out parent))
                continue;

            // create joint
            Specific_AddJoint(name, child, parent);
        }

        AddPhysicsSetupScript(gameObject, ragdollFile);

        if (!isKinematic)
            Utils.SetRenderersToUpdateWhenOffScreen(gameObject);
    }

    /// <summary>
    /// Add startup script to game object that will
    /// make each of it's own colliders ignore each other.
    /// Otherwise overlapping colliders will break the physics
    /// </summary>
    private static void AddPhysicsSetupScript(GameObject gameObject, RagdollFile ragdollFile)
    {
        var container = new RagdollBoneIgnoreGroups() { Groups = new RagdollBoneIgnoreGroup[ragdollFile.RagdollBones.Count - 1] };

        // create ignore groups
        for (int i = 0; i < ragdollFile.RagdollBones.Count - 1; i++)
        {
            var group = new RagdollBoneIgnoreGroup()
            {
                BoneName = ragdollFile.RagdollBones[i].BoneName,
                IgnoreBoneNames = new string[ragdollFile.RagdollBones.Count - (i + 1)]
            };

            // ignore all bones after the current one
            int index = 0;
            for (int j = i + 1; j < ragdollFile.RagdollBones.Count; j++)
            {
                group.IgnoreBoneNames[index] = ragdollFile.RagdollBones[j].BoneName;
                index++;
            }

            container.Groups[i] = group;
        }

        // add startup script, and add above structure as simple JSON
        var setup = gameObject.AddComponent<ModelPhysicsSetup>();
        setup.RigidBodyIgnoreGroupsJSON = EditorJsonUtility.ToJson(container);
    }

    /// <summary>
    /// Hooks up each material to the PBR materials
    /// generated in substance painter
    /// </summary>
    private static void SetupMaterials(GameObject gameObject, string assetPath, ModelImporter assetImporter)
    {
        // setup up material naming that doesn't conflict with other models,
        assetImporter.materialName = ModelImporterMaterialName.BasedOnModelNameAndMaterialName;

        // find PBR dir
        var pbrDir = Path.Combine(Path.GetDirectoryName(assetPath), "PBR");
        if (!Directory.Exists(pbrDir))
            return;

        // get renderer which has the material references
        var renderers = gameObject.transform.GetComponentsInChildren<Renderer>();
        if (renderers == null)
            return;

        foreach (var renderer in renderers)
        {
            foreach (var material in renderer.sharedMaterials)
            {
                // get prefix from material name
                string textureNamePrefix = material.name.Replace('-', '_');

                // main texture
                string albedoPath = Path.Combine(pbrDir, textureNamePrefix + "_AlbedoTransparency.png");
                var albedo = AssetDatabase.LoadAssetAtPath<Texture>(albedoPath);
                if (albedo != null)
                    material.SetTexture("_MainTex", albedo);

                // metallic texture
                string metallicPath = Path.Combine(pbrDir, textureNamePrefix + "_MetallicSmoothness.png");
                var metallic = AssetDatabase.LoadAssetAtPath<Texture>(metallicPath);
                if (metallic != null)
                    material.SetTexture("_MetallicGlossMap", metallic);

                // normal texture
                string normalPath = Path.Combine(pbrDir, textureNamePrefix + "_Normal.png");
                var normal = AssetDatabase.LoadAssetAtPath<Texture>(normalPath);
                if (normal != null)
                {
                    // ensure normal maps are imported with the right type
                    var texImporter = AssetImporter.GetAtPath(normalPath) as TextureImporter;
                    texImporter.textureType = TextureImporterType.NormalMap;

                    material.SetTexture("_BumpMap", normal);
                }

                // occlusion texture
                string occlusionPath = Path.Combine(pbrDir, textureNamePrefix + "_AO.png");
                var occlusion = AssetDatabase.LoadAssetAtPath<Texture>(occlusionPath);
                if (occlusion != null)
                    material.SetTexture("_OcclusionMap", occlusion);
            }
        }
    }

    #region Specific FBX files - setup

    private static bool Specific_IsStatic(string gameObjectName, string boneName, HashSet<string> debugLogHistory)
    {
        bool isStatic = false;

        switch (gameObjectName)
        {
            case "fence":
            case "rock":
            case "plane":
            case "tree":
            case "heli":
                isStatic = true;
                break;
            case "house":
                switch (boneName)
                {
                    case "hinged_door.door.bone1":
                    case "hinged_door.hinge.bone1":
                    case "hinged_door.hinge.bone12":
                        break;
                    default:
                        isStatic = true;
                        break;
                }
                break;
        }

        if (isStatic)
            DebugLog(debugLogHistory, "Marked as static, so given no rigidbody: {0}", gameObjectName);

        return isStatic;
    }

    private static bool Specific_BecomeRagdollOnCollision(string gameObjectName, HashSet<string> debugLogHistory)
    {
        bool canRagdoll = true;

        switch (gameObjectName)
        {
            case "octopus":
            case "fish":
                canRagdoll = false;
                break;
        }

        if (canRagdoll)
            DebugLog(debugLogHistory, "{0} will Ragdoll on collision", gameObjectName);

        return canRagdoll;
    }

    private static bool Specific_SetRigidBodyProperties(Rigidbody rigidBody, string name, string boneName)
    {
        // scale the mass
        rigidBody.mass *= 10;
        rigidBody.mass = Mathf.Max(rigidBody.mass, 10);

        // disable physics to begin with
        rigidBody.isKinematic = true;

        switch (name)
        {
            case "house":
                rigidBody.useGravity = false;
                rigidBody.drag = 8; // keep the player on the floor on contact

                switch (boneName)
                {
                    case "hinged_door.hinge.bone1":
                    case "hinged_door.hinge.bone12":
                        break;
                    default:
                        rigidBody.isKinematic = false;
                        break;
                }
                break;

            case "sword":
            case "shield":
            case "table":
                rigidBody.isKinematic = false;
                break;
        }

        return rigidBody.isKinematic;
    }

    /// <summary>
    /// Create joint between rigid bodies. Default to fixed joint,
    /// but we'll decide for other specific models to use different joints
    /// </summary>
    private static void Specific_AddJoint(string name, Rigidbody child, Rigidbody parent)
    {
        Joint joint = null;

        switch (name)
        {
            case "robot":
                joint = null;
                break;

            case "dino":
            case "stickman":
                joint = child.gameObject.AddComponent<CharacterJoint>();
                break;

            case "house":
                var hingeJoint = child.gameObject.AddComponent<HingeJoint>();
                hingeJoint.anchor = child.transform.InverseTransformPoint(parent.transform.position);
                hingeJoint.useLimits = true;
                hingeJoint.limits = new JointLimits() { min = -120, max = 120 };
                joint = hingeJoint;
                break;

            default:
                joint = child.gameObject.AddComponent<FixedJoint>();
                break;
        }

        if (joint != null)
            joint.connectedBody = parent;
    }

    /// <summary>
    /// For all static models we:
    /// 1. Turn off rigging, so a MeshRenderer is added instead of a SkinnedMeshRenderer (which isn't supported by lightmapping)
    /// 2. and tell Unity to generate UV coords from our UV map, for storing the models lightmap texture
    /// </summary>
    //private static void Specific_SetupLightmapping(string name, ModelImporter modelImporter, HashSet<string> debugLogHistory)
    //{
    //    switch (name)
    //    {
    //        case "plane":
    //        case "tree":
    //        case "rock":
    //        case "fence":
    //        case "track":
    //            modelImporter.animationType = ModelImporterAnimationType.None;
    //            modelImporter.generateSecondaryUV = true;
    //            DebugLog(debugLogHistory, "Skinned Renderer disabled for: " + name);
    //            break;
    //    }
    //}

    #endregion

    private static void DebugLog(HashSet<string> debugLogHistory, string format, params object[] args)
    {
        var message = string.Format(format, args);
        if (debugLogHistory.Contains(message))
            return;

        Debug.Log(message);
        debugLogHistory.Add(message);
    }
}

// -------------------------
// Serialised Types
// -------------------------

public class RagdollFile
{
    /// <summary>
    /// The ragdoll info per bone
    /// </summary>
    public List<RagdollBone> RagdollBones { get; set; }

    /// <summary>
    /// Generate a lookup using the bone name as the key
    /// </summary>
    public Dictionary<string, RagdollBone> GetLookup()
    {
        return RagdollBones.ToDictionary(
            b => b.BoneName,
            b => b);
    }
}

[XmlInclude(typeof(SphereRagdollBone))]
[XmlInclude(typeof(BoxRagdollBone))]
[XmlInclude(typeof(CapsuleRagdollBone))]
[XmlInclude(typeof(MeshRagdollBone))]
public abstract class RagdollBone
{
    /// <summary>
    /// Name of the bone
    /// </summary>
    public string BoneName { get; set; }
    /// <summary>
    /// Name of the parent bone
    /// </summary>
    public string ParentBoneName { get; set; }

    /// <summary>
    /// Unity Axes:    Y = Up, X = Right, Z = Forward
    /// Waveform Axes: Y = Up, X = Right, -Z = Forward
    /// 
    /// so we flip position on the X axis and rotation to match
    /// </summary>
    protected void ApplyColliderAxisFix(GameObject colliderContainer)
    {
        var pos = colliderContainer.transform.localPosition;
        pos.x *= -1;
        colliderContainer.transform.localPosition = pos;

        var rot = colliderContainer.transform.localEulerAngles;
        rot.y *= -1;
        rot.z *= -1;
        colliderContainer.transform.localEulerAngles = rot;
    }

    public abstract Collider AddCollider(GameObject colliderContainer, Rigidbody rigidBody, string assetDir);
}

public class SphereRagdollBone : RagdollBone
{
    /// <summary>
    /// The center relative to the bone position
    /// </summary>
    public Vec3 Center { get; set; }
    public float Radius { get; set; }
    public Quat Orientation { get; set; }

    public override Collider AddCollider(GameObject colliderContainer, Rigidbody rigidBody, string assetDir)
    {
        var collider = colliderContainer.AddComponent<SphereCollider>();
        collider.radius = Radius;

        colliderContainer.transform.localPosition = Center.ToVector3();
        colliderContainer.transform.localRotation = Orientation.ToQuaternion();
        ApplyColliderAxisFix(colliderContainer);

        if (rigidBody != null)
            rigidBody.mass = Radius * Radius;

        return collider;
    }
}

public class BoxRagdollBone : RagdollBone
{
    /// <summary>
    /// The center relative to the bone position
    /// </summary>
    public Vec3 Center { get; set; }
    public Vec3 Size { get; set; }
    public Quat Orientation { get; set; }

    public override Collider AddCollider(GameObject colliderContainer, Rigidbody rigidBody, string assetDir)
    {
        var collider = colliderContainer.AddComponent<BoxCollider>();
        collider.size = Size.ToVector3();

        colliderContainer.transform.localPosition = Center.ToVector3();
        colliderContainer.transform.localRotation = Orientation.ToQuaternion();
        ApplyColliderAxisFix(colliderContainer);

        if (rigidBody != null)
            rigidBody.mass = Size.X * Size.Y * Size.Z;

        return collider;
    }
}

public class CapsuleRagdollBone : RagdollBone
{
    const int X_AXIS = 0;
    const int Y_AXIS = 1;
    const int Z_AXIS = 2;

    /// <summary>
    /// The center relative to the bone position
    /// </summary>
    public Vec3 Center { get; set; }
    public float Radius { get; set; }
    public float Height { get; set; }
    public Quat Orientation { get; set; }

    public override Collider AddCollider(GameObject colliderContainer, Rigidbody rigidBody, string assetDir)
    {
        var collider = colliderContainer.AddComponent<CapsuleCollider>();
        collider.direction = Y_AXIS;
        collider.radius = Radius;

        // Waveform exports the capsule length which doesnt include the cap lengths at either end,
        // we need this in Unity so add the radius twice for both caps
        collider.height = Height + (Radius * 2);

        colliderContainer.transform.localPosition = Center.ToVector3();
        colliderContainer.transform.localRotation = Orientation.ToQuaternion();
        ApplyColliderAxisFix(colliderContainer);

        if (rigidBody != null)
            rigidBody.mass = Radius * Radius * Height;

        return collider;
    }
}

public class MeshRagdollBone : RagdollBone
{
    /// <summary>
    /// This is the relative path to the .obj
    /// </summary>
    public string CustomFileName { get; set; }
    /// <summary>
    /// When true, physics calculations will be faster
    /// </summary>
    public bool IsConvex { get; set; }

    public override Collider AddCollider(GameObject colliderContainer, Rigidbody rigidBody, string assetDir)
    {
        var asset = AssetDatabase.LoadAssetAtPath<Mesh>(Path.Combine(assetDir, CustomFileName));

        var collider = colliderContainer.AddComponent<MeshCollider>();
        collider.sharedMesh = asset;
        collider.convex = IsConvex;

        colliderContainer.transform.localPosition = Vector3.zero;
        colliderContainer.transform.localRotation = Quaternion.identity;

        if (rigidBody != null)
        {
            var size = asset.bounds.size;
            rigidBody.mass = size.x * size.y * size.z;
        }

        return collider;
    }
}

public class Vec3
{
    public float X { get; set; }
    public float Y { get; set; }
    public float Z { get; set; }

    public Vector3 ToVector3()
    {
        return new Vector3(X, Y, Z);
    }
}

public class Quat
{
    public float X { get; set; }
    public float Y { get; set; }
    public float Z { get; set; }
    public float W { get; set; }

    public Quaternion ToQuaternion()
    {
        return new Quaternion(X, Y, Z, W);
    }
}

[Serializable]
public class RagdollBoneIgnoreGroups
{
    public RagdollBoneIgnoreGroup[] Groups;
}

[Serializable]
public class RagdollBoneIgnoreGroup
{
    public string BoneName;
    public string[] IgnoreBoneNames;
}