﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml.Serialization;
using UnityEngine;

namespace Assets
{
    public static class Utils
    {
        /// <summary>
        /// returns new state
        /// </summary>
        public static bool ToggleIsKinematic(GameObject gameObject)
        {
            bool physicsIsEnabled = false;

            foreach (Rigidbody body in GetRigidBodies(gameObject))
            {
                body.isKinematic = !body.isKinematic;
                physicsIsEnabled = !body.isKinematic;
            }

            return physicsIsEnabled;
        }

        /// <summary>
        /// Gets the mesh's bones, in depth first order
        /// </summary>
        public static IEnumerable<Transform> GetBones(GameObject gameObject)
        {
            var stack = new Stack<Transform>(10);

            // push root_bone
            stack.Push(gameObject.transform.Find("root_bone"));

            do
            {
                var currentTransform = stack.Pop();

                // we'll process the child bones next
                foreach (Transform childTransform in currentTransform)
                {
                    if (childTransform.name != "Collider Container")
                        stack.Push(childTransform);
                }

                yield return currentTransform;
            }
            while (stack.Count > 0);
        }

        /// <summary>
        /// Gets the mesh's colliders, in depth first order
        /// </summary>
        public static IEnumerable<Collider> GetColliders(GameObject gameObject)
        {
            foreach (var bone in GetBones(gameObject))
            {
                if (bone.transform.childCount < 1)
                    continue;

                var colliderContainer = bone.GetChild(0);
                if (colliderContainer.name != "Collider Container")
                    continue;

                var collider = colliderContainer.GetComponent<Collider>();
                if (collider != null)
                    yield return collider;
            }
        }

        /// <summary>
        /// Gets the mesh's rigid bodies, in depth first order
        /// </summary>
        public static IEnumerable<Rigidbody> GetRigidBodies(GameObject gameObject)
        {
            foreach (var bone in GetBones(gameObject))
            {
                var rigidBody = bone.GetComponent<Rigidbody>();
                if (rigidBody != null)
                    yield return rigidBody;
            }
        }

        /// <summary>
        /// unity bug - need to turn this on otherwise mesh  
        /// gets culled even though we are looking at it when rigid body is active
        /// </summary>
        public static void SetRenderersToUpdateWhenOffScreen(GameObject gameObject)
        {
            foreach (var renderer in gameObject.GetComponentsInChildren<SkinnedMeshRenderer>())
            {
                renderer.updateWhenOffscreen = true;
            }
        }

        /// <summary>
        /// Load an XML file from the specified path.
        /// On error, return the default value for the specified type
        /// </summary>
        public static T LoadXml<T>(string filepath, Type[] extraTypes = null)
        {
            if (!File.Exists(filepath))
                return default(T);

            FileStream fs = null;
            try
            {
                fs = new FileStream(filepath, System.IO.FileMode.Open);
                var xs = new XmlSerializer(typeof(T), extraTypes);

                T data = (T)xs.Deserialize(fs);

                fs.Close();
                return data;
            }
            catch (Exception ex)
            {
                string error = ex.ToString();
                Debug.LogError(error);
            }
            finally
            {
                if (fs != null)
                    fs.Close();
            }

            return default(T);
        }
    }
}